//: [Previous](@previous)

import Foundation

class StateMachine {
    enum State: String, Equatable, CaseIterable {
        case none
        case loading
        case cancelling
        case error
        case empty
        case data
    }

    enum Event: String, Equatable, CaseIterable {
        case loadStarted
        case cancelStarted
        case cancelFinished
        case loadFailed
        case loadFinishedWithData
        case loadFinishedWithNoData
    }

    var state: State = .none {
        didSet {
            guard oldValue != state else { return }
        }
    }

    func transition(with event: Event) {
        switch (state, event) {
        case (.none, .loadStarted):
            state = .loading
        case (.none, _):
            break

        case (.loading, .cancelStarted):
            state = .cancelling
        case (.loading, .loadFailed):
            state = .error
        case (.loading, .loadFinishedWithNoData):
            state = .empty
        case (.loading, .loadFinishedWithData):
            state = .data
        case (.loading, _):
            break

        case (.cancelling, .cancelFinished):
            state = .none
        case (.cancelling, _):
            break

        case (.error, .loadStarted):
            state = .loading
        case (.error, _):
            break

        case (.empty, .loadStarted):
            state = .loading
        case (.empty, _):
            break

        case (.data, .loadStarted):
            state = .loading
        case (.data, _):
            break
        }
    }
}

let someDocWithMermaid = """
State machine models the data loading process for ....

```mermaid
graph LR
none --loadStarted--> loading

loading --cancelStarted--> cancelling
loading --loadFailed--> error
loading --loadFinishedWithData--> data
loading --loadFinishedWithNoData--> empty

cancelling --cancelFinished--> none
error --loadStarted--> loading

empty --loadStarted--> loading

data --loadStarted--> loading

```
"""

struct StateMachineDocumentation {
    struct TransitionDocumentation {
        var stateFrom: String
        var transition: String
        var stateTo: String
    }
    var transitions: [TransitionDocumentation]
}

extension StateMachineDocumentation {
    static let mermaidRegexp = try! NSRegularExpression(
        pattern: #"```mermaid(.|\n)*```"#,
        options: []
    )

    static func parseDocumentations(from document: String) -> [StateMachineDocumentation] {
        let matches = mermaidRegexp.matches(
            in: document, options: [],
            range: NSRange(location: 0, length: document.count)
        )

        return matches.map { .init(document: document, match: $0) }
    }

    init(document: String, match: NSTextCheckingResult) {
        self.transitions = TransitionDocumentation.parse(from: document, range: match.range)
    }
}

extension StateMachineDocumentation.TransitionDocumentation {
    static private let transitionRegexp = try! NSRegularExpression(
        pattern: #"^(\S*)\s*--\s*(\S*)-->\s*(\S*)$"#,
        options: [.anchorsMatchLines]
    )

    static func parse(from document: String, range: NSRange) -> [StateMachineDocumentation.TransitionDocumentation] {
        let matches = transitionRegexp.matches(
            in: document, options: [],
            range: range
        )

        return matches.compactMap { StateMachineDocumentation.TransitionDocumentation(document: document, match: $0) }
    }

    init?(document: String, match: NSTextCheckingResult) {
        guard
            match.numberOfRanges == 4,
            let stateFrom  = document.substring(nsRange: match.range(at: 1)),
            let transition = document.substring(nsRange: match.range(at: 2)),
            let stateTo    = document.substring(nsRange: match.range(at: 3))
        else {
            return nil
        }

        self.stateFrom = String(stateFrom)
        self.transition = String(transition)
        self.stateTo = String(stateTo)
    }
}

func testStateMachineTransitions() {
    // GIVEN
    let stateMachine = StateMachine()
    let stateMachineDoc = StateMachineDocumentation.parseDocumentations(from: someDocWithMermaid).first!

    for transitionDoc in stateMachineDoc.transitions {
        guard let stateFrom = StateMachine.State(rawValue: transitionDoc.stateFrom),
              let transition = StateMachine.Event(rawValue: transitionDoc.transition),
              let stateTo = StateMachine.State(rawValue: transitionDoc.stateTo) else {
            continue
        }

        stateMachine.state = stateFrom

        // WHEN
        stateMachine.transition(with: transition)

        // THEN
        assert(stateMachine.state == stateTo)
    }
}

testStateMachineTransitions()
